/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"

#include "../config.h"

#ifndef OPENCV_VER2
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cxcore.h>
#include <opencv/cvaux.h>
#endif

using namespace std;
using namespace FD;

class CvfdhCorner;
   
DECLARE_NODE(CvfdhCorner)
/*Node
 *
 * @name CvfdhCorner
 * @category FDOPENCV_HARK:Filter
 * @description CvfdhCorner
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @parameter_name CORNER_TYPE
 * @parameter_type string
 * @parameter_value EigenVal
 * @parameter_list EigenVal:Harris:FAST
 * @parameter_description Corner detection method selection
 *
 END*/
   
   
class CvfdhCorner : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
  string corner_type;
      
public:
  CvfdhCorner(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    corner_type = object_cast<String>(parameters.get("CORNER_TYPE"));
    inOrder = true;

#ifndef OPENCV_VER2
    if(corner_type == "FAST")
      cout << getName() << " : OpenCV less than Ver2.0 does not support FAST. Changed to Harris" << endl;    
#endif

  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {


    ObjectRef in = getInput(m_imageInID, count);

    // -------------------------------------------------------------

    if (typeid(*in) == typeid(CvfdImage)) {


      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	(*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
	return;
      }

      IplImage* img_in;
      img_in = cvCloneImage(imagePtr->getImage());

      IplImage* img_out;
    
#if ENABLE_OPENCV_2_0_0

      cv::Mat eigen_img(img_in);
    
      cv::Mat harris_img = eigen_img.clone();
      cv::Mat fast_img = eigen_img.clone();
    
      cv::Mat gray_img;
      cvtColor(eigen_img, gray_img, CV_BGR2GRAY);
      normalize(gray_img, gray_img, 0, 255, cv::NORM_MINMAX);
    
      vector<cv::Point2f> corners;
      vector<cv::Point2f>::iterator it_corner;

      if(corner_type == "EigenVal"){
      
	goodFeaturesToTrack(gray_img, corners, 80, 0.01, 5);
	it_corner = corners.begin();
	for(; it_corner!=corners.end(); ++it_corner) {
	  circle(eigen_img, cv::Point(it_corner->x, it_corner->y), 1, cv::Scalar(0,200,255), -1);
	  circle(eigen_img, cv::Point(it_corner->x, it_corner->y), 8, cv::Scalar(0,200,255));
	}
      
      }
    
      if(corner_type == "Harris"){
      
	goodFeaturesToTrack(gray_img, corners, 80, 0.01, 3, cv::Mat(), 3, true);
	it_corner = corners.begin();
	for(; it_corner!=corners.end(); ++it_corner) {
	  circle(harris_img, cv::Point(it_corner->x, it_corner->y), 1, cv::Scalar(0,255,0), -1);
	  circle(harris_img, cv::Point(it_corner->x, it_corner->y), 8, cv::Scalar(0,255,0 ));
	}

      }
  
      if(corner_type == "FAST"){
      
	int threshold = 100;
	bool nonmax = true;
	vector<cv::KeyPoint> keypoints;
	FAST(gray_img, keypoints, threshold, nonmax);
	vector<cv::KeyPoint>::iterator it_kp = keypoints.begin();
	for(; it_kp!=keypoints.end(); ++it_kp) {
	  circle(fast_img, cv::Point(it_kp->pt.x, it_kp->pt.y), 1, cv::Scalar(50,0,255), -1);
	  circle(fast_img, cv::Point(it_kp->pt.x, it_kp->pt.y), 8, cv::Scalar(50,0,255));
	}
      
      }

      img_out = cvCloneImage (img_in);


#else

      int i, corner_count;
      IplImage *dst_img, *src_img_gray;
      IplImage *eig_img, *temp_img;
      CvPoint2D32f *corners;
    
      dst_img = cvCloneImage (img_in);
      src_img_gray = cvCreateImage( cvGetSize(img_in),IPL_DEPTH_8U,1);
      cvCvtColor(img_in, src_img_gray, CV_BGR2GRAY);  
      eig_img = cvCreateImage (cvGetSize (src_img_gray), IPL_DEPTH_32F, 1);
      temp_img = cvCreateImage (cvGetSize (src_img_gray), IPL_DEPTH_32F, 1);
      corners = (CvPoint2D32f *) cvAlloc (corner_count * sizeof (CvPoint2D32f));
    
      if(corner_type == "EigenVal"){
	corner_count = 150;
	cvGoodFeaturesToTrack (src_img_gray, eig_img, temp_img, corners, &corner_count, 0.1, 15);
	cvFindCornerSubPix (src_img_gray, corners, corner_count,
			    cvSize (3, 3), cvSize (-1, -1), cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, 0.03));
      }
    
      if(corner_type == "Harris"){
	corner_count = 150;
	cvGoodFeaturesToTrack (src_img_gray, eig_img, temp_img, corners, &corner_count, 0.1, 15, NULL, 3, 1, 0.01);
	cvFindCornerSubPix (src_img_gray, corners, corner_count,
			    cvSize (3, 3), cvSize (-1, -1), cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, 0.03));
      }
    
      if(corner_type == "FAST"){
	corner_count = 150;
	cvGoodFeaturesToTrack (src_img_gray, eig_img, temp_img, corners, &corner_count, 0.1, 15, NULL, 3, 1, 0.01);
	cvFindCornerSubPix (src_img_gray, corners, corner_count,
			    cvSize (3, 3), cvSize (-1, -1), cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, 0.03));
      }
    
      for (i = 0; i < corner_count; i++)
	cvCircle (dst_img, cvPointFrom32f (corners[i]), 3, CV_RGB (0, 0, 255), 2);
    
      img_out = cvCloneImage (dst_img);    

      cvReleaseImage(&dst_img);   
      cvReleaseImage(&src_img_gray);
      cvReleaseImage(&eig_img);
      cvReleaseImage(&temp_img);

#endif
         
      CvfdImage* image= new CvfdImage(img_out);  
         
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage(&img_in);   
      cvReleaseImage(&img_out);   

      // -------------------------------------------------------------

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	IplImage* img_in;
	img_in = cvCloneImage(imagePtr.getImage());
	IplImage* img_out;

#if ENABLE_OPENCV_2_0_0

	cv::Mat eigen_img(img_in);
    
	cv::Mat harris_img = eigen_img.clone();
	cv::Mat fast_img = eigen_img.clone();
    
	cv::Mat gray_img;
	cvtColor(eigen_img, gray_img, CV_BGR2GRAY);
	normalize(gray_img, gray_img, 0, 255, cv::NORM_MINMAX);
    
	vector<cv::Point2f> corners;
	vector<cv::Point2f>::iterator it_corner;

	if(corner_type == "EigenVal"){
      
	  goodFeaturesToTrack(gray_img, corners, 80, 0.01, 5);
	  it_corner = corners.begin();
	  for(; it_corner!=corners.end(); ++it_corner) {
	    circle(eigen_img, cv::Point(it_corner->x, it_corner->y), 1, cv::Scalar(0,200,255), -1);
	    circle(eigen_img, cv::Point(it_corner->x, it_corner->y), 8, cv::Scalar(0,200,255));
	  }
      
	}
    
	if(corner_type == "Harris"){
      
	  goodFeaturesToTrack(gray_img, corners, 80, 0.01, 3, cv::Mat(), 3, true);
	  it_corner = corners.begin();
	  for(; it_corner!=corners.end(); ++it_corner) {
	    circle(harris_img, cv::Point(it_corner->x, it_corner->y), 1, cv::Scalar(0,255,0), -1);
	    circle(harris_img, cv::Point(it_corner->x, it_corner->y), 8, cv::Scalar(0,255,0 ));
	  }

	}
  
	if(corner_type == "FAST"){
      
	  int threshold = 100;
	  bool nonmax = true;
	  vector<cv::KeyPoint> keypoints;
	  FAST(gray_img, keypoints, threshold, nonmax);
	  vector<cv::KeyPoint>::iterator it_kp = keypoints.begin();
	  for(; it_kp!=keypoints.end(); ++it_kp) {
	    circle(fast_img, cv::Point(it_kp->pt.x, it_kp->pt.y), 1, cv::Scalar(50,0,255), -1);
	    circle(fast_img, cv::Point(it_kp->pt.x, it_kp->pt.y), 8, cv::Scalar(50,0,255));
	  }
      
	}

	img_out = cvCloneImage (img_in);

#else

	int i, corner_count;
	IplImage *dst_img, *src_img_gray;
	IplImage *eig_img, *temp_img;
	CvPoint2D32f *corners;
    
	dst_img = cvCloneImage (img_in);
	src_img_gray = cvCreateImage( cvGetSize(img_in),IPL_DEPTH_8U,1);
	cvCvtColor(img_in, src_img_gray, CV_BGR2GRAY);  
	eig_img = cvCreateImage (cvGetSize (src_img_gray), IPL_DEPTH_32F, 1);
	temp_img = cvCreateImage (cvGetSize (src_img_gray), IPL_DEPTH_32F, 1);
	corners = (CvPoint2D32f *) cvAlloc (corner_count * sizeof (CvPoint2D32f));
    
	if(corner_type == "EigenVal"){
	  corner_count = 150;
	  cvGoodFeaturesToTrack (src_img_gray, eig_img, temp_img, corners, &corner_count, 0.1, 15);
	  cvFindCornerSubPix (src_img_gray, corners, corner_count,
			      cvSize (3, 3), cvSize (-1, -1), cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, 0.03));
	}
    
	if(corner_type == "Harris"){
	  corner_count = 150;
	  cvGoodFeaturesToTrack (src_img_gray, eig_img, temp_img, corners, &corner_count, 0.1, 15, NULL, 3, 1, 0.01);
	  cvFindCornerSubPix (src_img_gray, corners, corner_count,
			      cvSize (3, 3), cvSize (-1, -1), cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, 0.03));
	}
    
	if(corner_type == "FAST"){
	  corner_count = 150;
	  cvGoodFeaturesToTrack (src_img_gray, eig_img, temp_img, corners, &corner_count, 0.1, 15, NULL, 3, 1, 0.01);
	  cvFindCornerSubPix (src_img_gray, corners, corner_count,
			      cvSize (3, 3), cvSize (-1, -1), cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, 0.03));
	}
    
	for (i = 0; i < corner_count; i++)
	  cvCircle (dst_img, cvPointFrom32f (corners[i]), 3, CV_RGB (0, 0, 255), 2);
    
	img_out = cvCloneImage (dst_img);    

	cvReleaseImage(&dst_img);   
	cvReleaseImage(&src_img_gray);
	cvReleaseImage(&eig_img);
	cvReleaseImage(&temp_img);

#endif

	CvfdImage* image= new CvfdImage(img_out);  
	(*outputMapPtr)[it->first] = image;

	cvReleaseImage(&img_in);   
	cvReleaseImage(&img_out);

      }
      (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
    }

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhCorner)
};
